@@ -3,7 +3,7 @@
  *
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
- * file, You can obtain one at http://mozilla.org/MPL/2.0/.
+ * file, you can obtain one at https://mozilla.org/MPL/2.0/.
  *
  * See the COPYRIGHT file distributed with this work for additional
  * information regarding copyright ownership.
@@ -143,6 +143,7 @@ EXTERN bool named_g_memstatistics INIT(false);
 EXTERN bool named_g_keepstderr	  INIT(false);
 
 EXTERN unsigned int named_g_tat_interval INIT(24 * 3600);
+EXTERN unsigned int named_g_maxcachesize INIT(0);
 
 #if defined(HAVE_GEOIP2)
 EXTERN dns_geoip_databases_t *named_g_geoip INIT(NULL);
